package ru.dbotthepony.mc.otm.android

import com.google.common.collect.ImmutableList
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.ComponentContents
import net.minecraft.network.chat.MutableComponent
import net.minecraft.network.chat.contents.TranslatableContents
import ru.dbotthepony.mc.otm.TranslatableComponent
import ru.dbotthepony.mc.otm.capability.android.AndroidCapabilityPlayer
import ru.dbotthepony.mc.otm.getKeyNullable
import ru.dbotthepony.mc.otm.registry.MRegistry

fun interface AndroidResearchFactory<R : AndroidResearch> {
	fun factory(type: AndroidResearchType<*>, capability: AndroidCapabilityPlayer): R
}

open class AndroidResearchType<R : AndroidResearch>(
	protected val factory: AndroidResearchFactory<R>
) {
	open val prerequisites: List<AndroidResearchType<*>> get() = emptyList()
	open val blockedBy: List<AndroidResearchType<*>> get() = emptyList()

	val blocking: List<AndroidResearchType<*>> by lazy {
		check(IS_REGISTRY_INITIALIZED) { "Trying to build list too early!" }

		val list = ImmutableList.builder<AndroidResearchType<*>>()

		for (research in MRegistry.ANDROID_RESEARCH) {
			if (research.blockedBy.contains(this)) {
				list.add(research)
			}
		}

		return@lazy list.build()
	}

	val unlocks: List<AndroidResearchType<*>> by lazy {
		check(IS_REGISTRY_INITIALIZED) { "Trying to build list too early!" }

		val list = ImmutableList.builder<AndroidResearchType<*>>()

		for (research in MRegistry.ANDROID_RESEARCH) {
			if (research.prerequisites.contains(this)) {
				list.add(research)
			}
		}

		return@lazy list.build()
	}

	fun factory(capability: AndroidCapabilityPlayer) = factory.factory(this, capability)

	val registryName by lazy {
		check(IS_REGISTRY_INITIALIZED) { "Trying to access registryName too early!" }
		MRegistry.ANDROID_RESEARCH.getKeyNullable(this)
	}

	val displayId by lazy {
		val registryName = registryName ?: throw NullPointerException("No registry name present")
		return@lazy "android_research.${registryName.namespace}.${registryName.path}".intern()
	}

	val descriptionId by lazy {
		return@lazy "$displayId.description".intern()
	}

	open val displayContents: ComponentContents by lazy {
		TranslatableContents(displayId)
	}

	open val displayName: Component by lazy {
		MutableComponent.create(displayContents)
	}

	open val displayDescription: List<Component> by lazy {
		listOf(TranslatableComponent(descriptionId))
	}

	companion object {
		var IS_REGISTRY_INITIALIZED = false
			private set

		fun markRegistryLoaded() {
			IS_REGISTRY_INITIALIZED = true
		}

		fun assert() {
			check(IS_REGISTRY_INITIALIZED) { "Registry is not yet initialized" }
		}

		inline fun assert(lazy: () -> Any) {
			check(IS_REGISTRY_INITIALIZED, lazy)
		}
	}
}
